# Python modules
import unittest

# 3rd Party Modules
import numpy as np

# Project modules
from src.gmix2 import gmix2

np.set_printoptions(precision=12)


class TestGmix2(unittest.TestCase):
    rng = np.random.RandomState(31)

    w = np.array([
        [0.15, 0.10, 0.40, 0.05, 0.3],
        [0.05, 0.20, 0.10, 0.35, 0.3]
    ])
    m = np.array([
        [ 0.0,  1.0, 2.0, 3.0, 4.0],
        [-5.0, -2.0, 0.0, 1.0, 3.0]
    ])
    s = np.array([
        [0.9, 0.5, 2.0, 1.0, 3.0],
        [0.5, 10.0, 2.0, 5.0, 0.8]
    ])

    def test_pdf(self):
        # One value at a time
        exp_pdf = np.array([0.212963552378, 0.059730094961])
        act_pdf = gmix2.pdf(1, self.w, self.m, self.s)

        np.testing.assert_allclose(act_pdf, exp_pdf)

        exp_pdf = [[0.212963552378, 0.000335617249],
                   [0.059730094961, 0.044152358959]]
        act_pdf = gmix2.pdf(np.array([1, -5.5]), self.w, self.m, self.s)

        np.testing.assert_allclose(act_pdf, exp_pdf)

        act_pdf = gmix2.pdf([1, -5.5], self.w, self.m, self.s)

        np.testing.assert_allclose(act_pdf, exp_pdf)

    def test_cdf(self):
        exp_cdf = [0.000266662424, 0.11474477839]
        act_cdf = gmix2.cdf(-5.5, self.w, self.m, self.s)

        np.testing.assert_allclose(act_cdf, exp_cdf)

        exp_cdf = [[0.352160058832, 0.000266662424],
                   [0.419591430163, 0.11474477839]]
        act_cdf = gmix2.cdf(np.array([1, -5.5]), self.w, self.m, self.s)

        np.testing.assert_allclose(act_cdf, exp_cdf)

        act_cdf = gmix2.cdf([1, -5.5], self.w, self.m, self.s)

        np.testing.assert_allclose(act_cdf, exp_cdf)

    def test_ppf(self):
        exp_ppf = [0.753339733752, -1.185816277994]
        act_ppf = gmix2.ppf(0.3, self.w, self.m, self.s)

        np.testing.assert_allclose(act_ppf, exp_ppf)

        exp_ppf = [[0.753339733752, 5.673110996856],
                   [-1.185816277994, 5.916783093835]]
        act_ppf = gmix2.ppf(np.array([0.3, 0.9]), self.w, self.m, self.s)

        np.testing.assert_allclose(act_ppf, exp_ppf)

        act_ppf = gmix2.ppf([0.3, 0.9], self.w, self.m, self.s)

        np.testing.assert_allclose(act_ppf, exp_ppf)

    def test_mean(self):
        exp_mean = [2.25, 0.6]
        act_mean = gmix2.mean(self.w, self.m)

        np.testing.assert_allclose(act_mean, exp_mean)

    def test_median(self):
        exp_median = [1.79712881, 2.00299055]
        act_median = gmix2.median(self.w, self.m, self.s)

        np.testing.assert_allclose(act_median, exp_median)

    def test_mode(self):
        exp_mode = [[(0.9580152000982137, 0.21324093573952826)],
                    [(-4.970966372242318, 0.06205794913899075),
                     (2.96872434469676, 0.189013117805071)]]
        act_mode = gmix2.mode(self.w, self.m, self.s)

        for act_grp, exp_grp in zip(act_mode, exp_mode):
            for (act_x, act_p), (exp_x, exp_p) in zip(act_grp, exp_grp):
                self.assertAlmostEqual(act_x, exp_x)
                self.assertAlmostEqual(act_p, exp_p)

    def test_variance(self):
        exp_var = [6.384, 34.0945]
        act_var = gmix2.variance(self.w, self.m, self.s)

        np.testing.assert_allclose(act_var, exp_var)



import cgmix2

import logging
import operator as op

import numpy as np

from cgmix2 import *
from functools import reduce
from typing import List, Optional, Tuple

log = logging.getLogger(__name__)


def _make_mixture_names(names, n):
    names = np.array(names) if names else np.arange(n)
    return (
        np.char.mod('%03d', names)
        if all([isinstance(n, (int, np.int, np.int16, np.int32, np.int64)) for n in names])
        else names
    )


def _make_plot_range(x_min, x_max, mu, sigma):
    min_mu, max_mu, max_sigma = np.min(mu), np.max(mu), np.max(sigma)

    x_min = min_mu - 2.5 * max_sigma if x_min is None else x_min
    x_max = max_mu + 2.5 * max_sigma if x_max is None else x_max

    return x_min, x_max


def _make_mixture_parameters(weight, mean, std):
    import pandas as pd

    # A simple utility method for making a data frame for each parameter array
    # of the correct shape.
    def make_df(a: np.ndarray, value_name: str):
        t = pd.DataFrame(a.transpose())
        t.columns = np.arange(a.shape[0])
        t['kernel'] = np.arange(a.shape[1])

        return t.melt('kernel', var_name='mixture_name', value_name=value_name)

    # Merge each individual dataframe into a single one.
    dfs = (make_df(weight, 'weight'), make_df(mean, 'mu'), make_df(std, 'sigma'))
    return reduce(lambda a, b: pd.merge(a, b, how='left', on=['kernel', 'mixture_name']), dfs)


def _make_mixture_frame(weight, mean, std, x):
    import pandas as pd

    try:
        from scipy.stats import norm
    except ImportError:
        raise ImportError("You need scipy in order plot the individual mixtures.")

    param_df = _make_mixture_parameters(weight, mean, std)

    # Compute the weighted Normal pdf values for each row in the param_df and
    # store them in a new data frame.
    normal_df = pd.DataFrame(
        np.repeat(param_df.values, len(x), axis=0),
        columns=param_df.columns,
        index=np.tile(x, len(param_df))
    ).rename_axis('x').reset_index()

    # Calculate the pdf values for each data point
    normal_df['y'] = normal_df.apply(
        lambda r: r['weight'] * norm.pdf(r['x'], r['mu'], r['sigma']),
        axis=1
    )

    # Convert the kernel and group columns to categorical data types
    cat_columns = ['kernel', 'mixture_name']
    normal_df[cat_columns] = normal_df[cat_columns].apply(lambda t: t.astype('category'))

    # Reorder the columns to put y next to x (for human readability)
    normal_df = normal_df[['mixture_name', 'kernel', 'weight', 'mu', 'sigma', 'x', 'y']]

    df = normal_df[['mixture_name', 'x', 'y']].groupby(
        ['mixture_name', 'x']
    ).agg({
        'y': 'sum'
    }).reset_index()

    return normal_df, df


def _make_point_df(x, weight, mean, std, mixture_names):
    import pandas as pd

    y = [
        pdf(xv.item(), weight[None, i], mean[None, i], std[None, i])[0]
        for i, xv in enumerate(np.nditer(x))
    ]
    df = pd.DataFrame({'x': x, 'y': y, 'mixture_name': mixture_names})

    return df


# noinspection PyTypeChecker
def plot(
        weights: np.ndarray,
        means: np.ndarray,
        stds: np.ndarray,
        mixture_names: Optional[List[str]] = None,
        x_min: Optional[float] = None,
        x_max: Optional[float] = None,
        x_ticks: Optional[List[float]] = None,
        n_points: int = 1000,
        show_mixture: bool = False,
        show_mean: bool = False,
        show_median: bool = False,
        show_mode: bool = False,
        percent_interval: Optional[float] = None,
        legend_title: str = 'Mixture Name',
        wrap_cols: Optional[int] = None,
        palette: str = 'Dark2',
        figure_size: Tuple[float, float] = None
):
    """
    Plot one or more mixtures. By default multiple mixtures will be shown in the same plotting area each with
    a separate color. You can use ``wrap_cols`` to plot them in on separate areas with ``wrap_cols`` plots
    per row.

    :param weights: The weights of the mixtures to plot. This should be a 2 dimensional shape where each column
           is an individual weight (that sums to 1.0) and each row is a different mixture.
    :param means: The means of the mixtures.
    :param stds: The standard deviations of the mixtures.
    :param mixture_names: The names of each mixture, if given, otherwise generic names will be provided.
    :param x_min: The minimum x value to plot. If no value is given a default value based on the distributions
           will be used.
    :param x_max: The maximum x value to plot. If no value is given a default value based on the distributions
           will be used.
    :param x_ticks: Explicitly provide breaks for the x-axis tick marks.
    :param n_points: The number of points to plot. More points will provide a smoother curve, but will take
           more time to plot.
    :param show_mixture: If ``True``, each individual component will also be shown on the plot.
    :param show_mean: If ``True``, the mean value will be shown on the plot as a text annotation.
    :param show_median: If ``True``, the median value will be shown on the plot as a text annotation.
    :param show_mode: If ``True``, the modes of the plot will be shown on the plot as a text annotation.
    :param percent_interval: If this is not ``None`` the given percent interval will be highlighted on the plot.
    :param legend_title: Optionally, give the legend a specific title other than the default.
    :param wrap_cols: Plot each mixture individually and wrap the figure at this many columns.
    :param palette: Choose a colorbrewer palette for plotting multiple mixtures.
    :param figure_size: Specify a figure size in inches.
    :return: The `plotnine <https://plotnine.readthedocs.io/en/stable/>`_ object.
    """
    try:
        import pandas as pd
        import plotnine as p9
        # NOTE: pycairo needs to be installed separately otherwise matplotlib will complain
        #       You also need to install a package for rendering (e.g., pip install PyQt5 works)
    except ImportError:
        raise ValueError("You need Pandas and Plotnine to plot distributions")

    if weights.ndim == 1:
        weights = np.expand_dims(weights, 0)
        means = np.expand_dims(means, 0)
        stds = np.expand_dims(stds, 0)

    n_mixtures = weights.shape[0]
    mixture_names = _make_mixture_names(mixture_names, n_mixtures)

    x_min, x_max = _make_plot_range(x_min, x_max, means, stds)
    x = np.linspace(x_min, x_max, num=n_points)

    show_legend = n_mixtures > 1

    p = p9.ggplot()

    if show_mixture:
        normal_df, df = _make_mixture_frame(weights, means, stds, x)
        df['mixture_name'] = np.repeat(mixture_names, n_points)

        p += p9.geom_path(
            data=normal_df,
            mapping=p9.aes('x', 'y', color='mixture_name'),
            alpha=0.5,
            show_legend=False
        )

    # If we aren't showing the individual mixtures we can compute the pdf
    # using the C++ interface.
    the_pdf = pdf(x, weights, means, stds).ravel()
    df = pd.DataFrame(the_pdf, columns=['y'])
    df['x'] = np.tile(x, n_mixtures)

    df['mixture_name'] = np.repeat(mixture_names, n_points)

    # Add some end points so that the area under the curve gets filled in
    # completely along the x-axis by the geom_polygon.
    end_point_data = []
    for grp in mixture_names:
        end_point_data.append({'mixture_name': grp, 'x': x_min - 0.1, 'y': 0})
        end_point_data.append({'mixture_name': grp, 'x': x_max + 0.1, 'y': 0})
    end_point_df = pd.DataFrame(end_point_data)
    df = df.append(end_point_df)
    df.sort_values(['mixture_name', 'x'], inplace=True)

    p += p9.geom_polygon(
        data=df,
        mapping=p9.aes('x', 'y', fill='mixture_name'),
        alpha=0.25,
        color='black',
        show_legend=show_legend
    )

    if show_mean:
        mean_df = _make_point_df(cgmix2.mean(weights, means), weights, means, stds, mixture_names)
        segment_aes = p9.aes(x='x', xend='x', y=0, yend='y', color='mixture_name')
        text_aes = p9.aes(x='x', y=0, label='x')
        text_kw = {'format_string': 'μ: {:0.3f}', 'ha': 'left', 'va': 'top'}

        p += p9.geom_segment(mean_df, segment_aes, size=1.25, show_legend=False)
        p += p9.geom_text(text_aes, data=mean_df, **text_kw)

    if show_median or percent_interval:
        median_df = _make_point_df(median(weights, means, stds), weights, means, stds, mixture_names)
        segment_aes = p9.aes(x='x', xend='x', y=0, yend='y', color='mixture_name')
        text_aes = p9.aes(x='x', y='y', label='x')
        text_kw = {'format_string': 'x̃: {:0.3f}', 'ha': 'left', 'va': 'bottom'}

        p += p9.geom_segment(
            median_df,
            segment_aes,
            linetype='dashed',
            size=1.25,
            show_legend=False
        )
        p += p9.geom_text(text_aes, data=median_df, **text_kw)

    if show_mode:
        modes = mode(weights, means, stds)

        d = [
            dict(mixture_name=mixture_names[i], x=a, y=b)
            for i, m in enumerate(modes)
            for a, b in m
        ]

        mode_df = pd.DataFrame(d)

        segment_aes = p9.aes(x='x', xend='x', y=0, yend='y', color='mixture_name')
        text_aes = p9.aes(x='x', y='y/2', label='x', angle=-90)
        text_kw = {'format_string': 'x̂: {:0.3f}', 'ha': 'left', 'va': 'center', 'nudge_y': -0.005}

        p += p9.geom_segment(
            mode_df,
            segment_aes,
            linetype='dotted',
            size=1.25,
            show_legend=False
        )
        p += p9.geom_text(text_aes, data=mode_df, **text_kw)

    if percent_interval:
        if not 0 < percent_interval < 1:
            log.error(f"The percent interval must be between 0 and 100: {percent_interval}")
            raise ValueError("The percent interval must be between 0 and 100")

        def make_limit_df(bound_df, is_lower: bool):
            fn = op.le if is_lower else op.ge

            start = -1 if is_lower else 0
            stop = None if is_lower else 1

            # Select all the rows from the gmix_df that are outside the PI
            # for each group
            df_list = [
                df.loc[(fn(df.x, row.x)) & (df.mixture_name == row.mixture_name), ]
                for _, row in bound_df.iterrows()
            ]
            tmp_df = pd.concat(df_list)

            # We need to drop the polygon to zero at the interval boundary,
            # so create a new row that has the same x values of the last point
            # of the lower bound and the first point of the upper bound, but
            # set the y value to 0.
            last_point = tmp_df.groupby('mixture_name').apply(
                lambda g: pd.DataFrame(g.iloc[start:stop, ].values, columns=g.columns)
            )
            last_point['y'] = 0
            tmp_df = tmp_df.append(last_point)
            tmp_df.reset_index(drop=True, inplace=True)

            # Make sure the columns are of the correct type (for some reason they
            # get reset to 'object' above.
            tmp_df['mixture_name'] = tmp_df.mixture_name.astype('category')
            tmp_df[['x', 'y']] = tmp_df[['x', 'y']].apply(lambda c: c.astype('float'))

            return tmp_df

        h = (1 - percent_interval) / 2
        lb = h
        ub = percent_interval + h

        lb_df = _make_point_df(ppf(lb, weights, means, stds), weights, means, stds, mixture_names)
        up_df = _make_point_df(ppf(ub, weights, means, stds), weights, means, stds, mixture_names)

        lower_df = make_limit_df(lb_df, True)
        upper_df = make_limit_df(up_df, False)

        poly_kw = {'alpha': 0.6, 'show_legend': False}

        p += p9.geom_polygon(lower_df, p9.aes('x', 'y', fill='mixture_name'), **poly_kw)
        p += p9.geom_polygon(upper_df, p9.aes('x', 'y', fill='mixture_name'), **poly_kw)

    if wrap_cols and n_mixtures > 1:
        p += p9.facet_wrap('~ mixture_name', ncol=wrap_cols)

    if x_ticks:
        p += p9.scale_x_continuous(breaks=x_ticks)

    if n_mixtures == 1:
        p += p9.scale_fill_grey()
        p += p9.scale_color_grey()

        title = f"{mixture_names[0]}"

        if percent_interval:
            title += f" With {percent_interval:0.3f} PI"
    else:
        p += p9.scale_fill_brewer(palette=palette, type='qualitative', name=legend_title)
        p += p9.scale_color_brewer(palette=palette, type='qualitative')

        title = "Density Plot"

        if percent_interval:
            title += f" With {percent_interval:0.3f} PI"

    p += p9.ylab('Density')
    p += p9.ggtitle(title)
    p += p9.theme_bw()

    if figure_size:
        p += p9.theme(figure_size=figure_size)

    return p



import os
import pathlib
import subprocess

from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext


current_dir = pathlib.Path(__file__).parent
readme = (current_dir / "README.rst").read_text()


# See: https://www.benjack.io/2017/06/12/python-cpp-tests.html
class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        super().__init__(name, sources=[])

        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: " +
                ", ".join(e.name for e in self.extensions)
            )

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        cwd = pathlib.Path().absolute()

        # these dirs will be created in build_py, so if you don't have
        # any python sources to bundle, the dirs will be missing
        build_temp = pathlib.Path(self.build_temp)
        build_temp.mkdir(parents=True, exist_ok=True)

        # FreeBSD won't build properly if the extension directory name is the same name as the module.
        extdir = pathlib.Path(self.get_ext_fullpath(f'ext_{ext.name}'))
        extdir.mkdir(parents=True, exist_ok=True)

        # example of cmake args
        config = 'Debug' if self.debug else 'Release'
        cmake_args = [
            '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + str(extdir.parent.absolute()),
            '-DCMAKE_BUILD_TYPE=' + config
        ]

        # example of build args
        build_args = [
            '-v',
            '--config', config,
            '--', '-j2'
        ]

        os.chdir(str(build_temp))
        self.spawn(['cmake', str(cwd)] + cmake_args)
        if not self.dry_run:
            self.spawn(['cmake', '--build', '.'] + build_args)
        os.chdir(str(cwd))


if __name__ == '__main__':
    setup(name='gmix2',
          version='0.1.7',
          description='An implementation of Gaussian mixture distribution methods',
          long_description=readme,
          long_description_content_type='text/x-rst',
          install_requires=[
              'numpy',
          ],
          extras_require={
              'plotting': [
                  'pandas',
                  'plotnine',
                  'scipy'
              ]
          },
          author='Reid Swanson',
          maintainer='Reid Swanson',
          author_email='reid@reidswanson.com',
          maintainer_email='reid@reidswanson.com',
          zip_safe=False,
          package_dir={'': 'src/gmix2'},
          py_modules=['gmix2'],
          ext_modules=[CMakeExtension('cgmix2')],
          cmdclass={'build_ext': CMakeBuild},
          include_package_data=True,
          license='Apache-2.0',
          url='https://bitbucket.org/reidswanson/gmix2',
          classifiers=['Development Status :: 4 - Beta',
                       'Intended Audience :: Science/Research',
                       'License :: OSI Approved :: Apache Software License',
                       'Natural Language :: English',
                       'Operating System :: POSIX :: Linux',
                       'Programming Language :: Python :: 3.7',
                       'Topic :: Scientific/Engineering :: Mathematics'])
